package main

import (
	"encoding/json"
	"encoding/xml"
	"fmt"
	"github.com/go-openapi/spec"
	"io/ioutil"
	"net/http"
	"os"
	"os/exec"
	"runtime"
	"sort"
	"strconv"
	"strings"
	"time"
)

type InitialMenu struct {
	Initial  string    `json:"initial"`
	Icon     string    `json:"icon"`
	DirMenus []DirMenu `json:"dirMenus"`
}
type DirMenu struct {
	Name     string     `json:"name"`
	SubMenus []spec.Tag `json:"subMenus"`
}

var InitialToIcon = map[string]string{
	"A": "el-icon-apple",
	"B": "el-icon-bicycle",
	"C": "el-icon-cherry",
	"D": "el-icon-dish",
	"E": "el-icon-eleme",
	"F": "el-icon-food",
	"G": "el-icon-grape",
	"H": "el-icon-house",
	"I": "el-icon-ice-cream-round",
	"J": "el-icon-dessert", //jam
	"K": "el-icon-key",
	"L": "el-icon-lock",
	"M": "el-icon-moon",
	"N": "el-icon-notebook-2",
	"O": "el-icon-orange",
	"P": "el-icon-pear",
	"Q": "el-icon-question",
	"R": "el-icon-refrigerator",
	"S": "el-icon-sunny",
	"T": "el-icon-trophy",
	"U": "el-icon-umbrella",
	"V": "el-icon-video-camera-solid",
	"W": "el-icon-watch",
	"X": "el-icon-close", //
	"Y": "el-icon-check", //yes
	"Z": "el-icon-zoom-in",
}

var InitialToMean = map[string]string{
	"A": "Apple",
	"B": "Bicycle",
	"C": "Cherry",
	"D": "Dish",
	"E": "Eleme",
	"F": "Food",
	"G": "Grape",
	"H": "House",
	"I": "IceCream",
	"J": "Jam", //jam
	"K": "Key",
	"L": "Lock",
	"M": "Moon",
	"N": "Notebook",
	"O": "Orange",
	"P": "Pear",
	"Q": "Question",
	"R": "Refrigerator",
	"S": "Sunny",
	"T": "Trophy",
	"U": "Umbrella",
	"V": "VideoCamera",
	"W": "Watch",
	"X": "X",   //
	"Y": "Yes", //yes
	"Z": "Zoom-in",
}

var NameToSwaggerBytes = make(map[string][]byte)

type InitialMenus []InitialMenu

func (p InitialMenus) Len() int {
	return len(p)
}

func (p InitialMenus) Swap(i, j int) {
	p[i], p[j] = p[j], p[i]
}

func (p InitialMenus) Less(i, j int) bool {
	return p[i].Initial < p[j].Initial
}

type DirMenus []DirMenu

func (p DirMenus) Len() int {
	return len(p)
}

func (p DirMenus) Swap(i, j int) {
	p[i], p[j] = p[j], p[i]
}

func (p DirMenus) Less(i, j int) bool {
	return p[i].Name < p[j].Name
}

type MyXmlApplication struct {
	xml *XmlApplication
	Dir string
}

func (app *MyXmlApplication) GetDir(controller XmlController) string {
	dir := app.Dir
	if app.Dir == "" {
		dir = controller.Dir
	}
	return dir
}

func GetMenu(apps []*MyXmlApplication, token string) []DirMenu {
	dirToTags := make(map[string][]spec.Tag)
	for _, app := range apps {
		for _, controller := range app.xml.Controllers.ControllerList {
			if token == "t" && controller.SkipLogin {
				continue
			} else if token == "s" && !controller.SkipLogin {
				continue
			}
			tag := spec.Tag{
				TagProps: spec.TagProps{
					Name: controller.Name,
					//Description: controller.Desc + ",Dir:" + controller.Dir + ",SkipLogin:" + strconv.FormatBool(controller.SkipLogin),
				},
			}
			dir := app.GetDir(controller)
			dirToTags[dir] = append(dirToTags[dir], tag)
		}
	}
	dirMenus := make([]DirMenu, 0)
	for dir, tags := range dirToTags {
		dirMenus = append(dirMenus, DirMenu{
			Name:     dir,
			SubMenus: tags,
		})
	}
	sort.Sort(DirMenus(dirMenus))

	return dirMenus
}
func GetTypeAndFormatAndDefault(propType string, defaultStr string) (string, string, interface{}) {
	_type := propType
	format := ""
	var defaultValue interface{}
	switch propType {
	case "int32":
		_type = "integer"
		format = propType
		if defaultStr != "" {
			defaultValue, _ = strconv.ParseInt(defaultStr, 10, 32)
		} else {
			defaultValue = 0
		}
	case "int64":
		_type = "integer"
		format = propType
		if defaultStr != "" {
			defaultValue, _ = strconv.ParseInt(defaultStr, 10, 32)
		} else {
			defaultValue = 0
		}
	case "local_time":
		_type = "string"
		format = "date-time"
	case "string":
		defaultValue = defaultStr
	}
	return _type, format, defaultValue
}
func GetSwaggerDefinitions(beans XmlBeans, tables XmlTables) spec.Definitions {
	definitions := spec.Definitions{}
	for _, bean := range beans.BeanList {
		desc := bean.Desc
		if bean.Inher != "" {
			desc += "| Inher:" + bean.Inher
		}
		if bean.XMLName.Local != "" {
			desc += "| xl:" + bean.XMLName.Local
		}
		//if bean.XMLName.Space != "" {
		//	desc += "| sp:" + bean.XMLName.Space
		//}
		properties := spec.SchemaProperties{}
		for _, prop := range bean.PropList {
			_type, format, defaultValue := GetTypeAndFormatAndDefault(prop.Type, prop.DefaultValue)
			properties[prop.Name] = spec.Schema{
				SchemaProps: spec.SchemaProps{
					Type:        spec.StringOrArray{_type},
					Format:      format,
					Description: prop.Caption,
					Title:       prop.Name,
					Default:     defaultValue,
				},
			}
		}
		definitions[bean.Name] = spec.Schema{
			SchemaProps: spec.SchemaProps{
				Type:        spec.StringOrArray{"object"},
				Description: desc,
				Title:       bean.Name,
				Properties:  properties,
			},
		}
	}
	for _, table := range tables.TableList {
		desc := table.Desc
		if table.XMLName.Local != "" {
			desc += "| xl:" + table.XMLName.Local
		}
		//if table.XMLName.Space != "" {
		//	desc += "| sp:" + table.XMLName.Space
		//}
		properties := spec.SchemaProperties{}
		for _, prop := range table.ColumnList {
			_type, format, defaultValue := GetTypeAndFormatAndDefault(prop.Type, prop.DefaultValue)
			properties[prop.Name] = spec.Schema{
				SchemaProps: spec.SchemaProps{
					Type:        spec.StringOrArray{_type},
					Format:      format,
					Description: prop.Caption,
					Title:       prop.Name,
					Default:     defaultValue,
				},
			}
		}
		definitions[table.Name] = spec.Schema{
			SchemaProps: spec.SchemaProps{
				Type:        spec.StringOrArray{"object"},
				Description: desc,
				Title:       table.Name,
				Properties:  properties,
			},
		}
	}
	return definitions
}
func GetSwagger(controller XmlController, allDefinitions spec.Definitions) *spec.Swagger {
	definitions := spec.Definitions{}
	swagger := spec.Swagger{}
	swagger.Swagger = "2.0"
	swagger.Info = &spec.Info{
		InfoProps: spec.InfoProps{Title: controller.Name},
	}
	swagger.Host = os.Getenv("SWAGGER_HOST")
	if swagger.Host == "" {
		swagger.Host = "localhost:8086"
	}
	swagger.BasePath = ""
	tagName := controller.Name
	swagger.Tags = []spec.Tag{{
		TagProps: spec.TagProps{
			Name:        tagName,
			Description: controller.Desc,
		},
	}}
	needToken := !controller.SkipLogin
	if needToken {
		swagger.SecurityDefinitions = spec.SecurityDefinitions{
			"token": &spec.SecurityScheme{
				SecuritySchemeProps: spec.SecuritySchemeProps{
					Type: "apiKey",
					Name: "token",
					In:   "header",
				},
			},
		}
	}
	swaggerScheme := os.Getenv("SWAGGER_SCHEME")
	if swaggerScheme != "https" {
		swaggerScheme = "http"
	}
	swagger.Schemes = []string{swaggerScheme}
	swagger.Paths = new(spec.Paths)
	swagger.Paths.Paths = make(map[string]spec.PathItem)
	tag := []string{tagName}
	for _, api := range controller.Apis {
		pathItem := new(spec.PathItem)
		operation := new(spec.Operation)
		operation.ID = api.Name
		operation.Summary = api.Desc
		operation.Description = "Function:" + api.Function + ", Table:" + api.Table
		//method
		methods := strings.Split(api.Method, ",")
		{
			for _, method := range methods {
				switch strings.ToLower(method) {
				case "get":
					pathItem.Get = operation
				case "put":
					pathItem.Put = operation
				case "post":
					pathItem.Post = operation
				case "delete":
					pathItem.Delete = operation
				case "options":
					pathItem.Options = operation
				case "head":
					pathItem.Head = operation
				case "patch":
					pathItem.Patch = operation
				}
			}
		}
		//params
		{
			params := make([]spec.Parameter, 1)
			p := spec.Parameter{}
			for _, method := range methods {
				switch strings.ToLower(method) {
				case "get":
					p.In = "path"
				default:
					p.In = "body"
					p.Name = "body"
				}
			}
			p.Required = false
			paramModel := "type"
			{
				if len(api.ParamList) == 1 {
					for _, param := range api.ParamList {
						if param.Type == "" && param.Ref != "" {
							paramModel = "ref"
						}
					}
				}
				if paramModel == "type" {
					for _, param := range api.ParamList {
						if param.Type == "" || param.Ref != "" {
							paramModel = ""
							break
						}
					}
				}
				if paramModel == "" {
					if len(api.ParamList) != 0 {
						paramModel = "unknown"
					}
				}
			}
			if paramModel == "type" {
				properties := spec.SchemaProperties{}
				for _, param := range api.ParamList {
					schema := spec.Schema{}
					schema.Title = param.Name
					schema.Description = param.Desc
					if param.TransType != "" {
						schema.Description += "|" + param.TransType
					}
					schema.Nullable = !param.Must
					_type, format, defaultValue := GetTypeAndFormatAndDefault(param.Type, param.DefaultValue)
					schema.Default = defaultValue
					schema.Type = spec.StringOrArray{_type}
					schema.Format = format
					//if param.Ref != "" {
					//	schema.Type = spec.StringOrArray{"object"}
					//	schema.Ref = spec.MustCreateRef("#/definitions/" + strings.Trim(param.Ref, "$"))
					//}
					properties[param.Name] = schema
				}
				p.Schema = &spec.Schema{}
				p.Schema.Properties = properties
				params[0] = p
				operation.Parameters = params
			} else if paramModel == "ref" {
				for _, param := range api.ParamList {
					schema := spec.Schema{}
					schema.Title = param.Name
					schema.Description = param.Desc
					if param.TransType != "" {
						schema.Description += "|" + param.TransType
					}
					schema.Nullable = !param.Must
					schema.Default = param.DefaultValue
					if param.Ref != "" {
						ref := strings.Trim(param.Ref, "$")
						_, exist := allDefinitions[ref]
						if exist {
							//schema.Type = spec.StringOrArray{"object"}
							schema.Ref = spec.MustCreateRef("#/definitions/" + ref)
							definitions[ref] = allDefinitions[ref]
						}
					}
					p.Schema = &schema
				}
				params[0] = p
				operation.Parameters = params
			} else if paramModel == "unknown" {
				params[0] = p
				operation.Parameters = params
			}
		}
		//return
		{
			operation.Responses = &spec.Responses{}
			success := spec.Response{}
			success.Description = api.Return.Success.Desc
			success.AddExtension("ref", api.Return.Success.Ref)
			failure := spec.Response{}
			failure.Description = api.Return.Failure.Desc
			failure.AddExtension("ref", api.Return.Failure.Ref)
			operation.Responses.StatusCodeResponses = map[int]spec.Response{
				200:  success,
				-200: failure,
			}
		}
		operation.Tags = tag
		if needToken {
			operation.Security = []map[string][]string{
				{"token": {}},
			}
		}
		swagger.Paths.Paths["/api/v1/"+controller.Name+"/"+api.Name] = *pathItem
	}
	swagger.Definitions = definitions
	return &swagger
}

func GetProjectApp() *MyXmlApplication {
	projName := "i2-erp-backend"
	app := NewEngineClient(projName, "http://gen.qianqiusoft.com").GenerateCurrentProject()
	if app == nil {
		app = &XmlApplication{}
		fileName := GetRightFile(
			[]string{
				os.Getenv("GOPATH") + "/src/i2-erp-backend/i2-erp-backend.proj/i2-erp-backend_temp.xml",
				"../i2-erp-backend/i2-erp-backend.proj/i2-erp-backend_temp.xml",
				"./i2-erp-backend.proj/i2-erp-backend_temp.xml",
				"./config/i2-erp-backend_temp.xml",
				"./i2-erp-backend_temp.xml",
			},
		)
		fmt.Println("read:" + fileName)
		bytess, _ := ioutil.ReadFile(fileName)
		err := xml.Unmarshal(bytess, app)
		if err != nil {
			fmt.Println("xml.Unmarshal(bytess, &app) error " + err.Error())
			return nil
		}
	}
	return &MyXmlApplication{xml: app}
}

func GetApiLightApp() *MyXmlApplication {
	app := &XmlApplication{}
	fileName := GetRightFile(
		[]string{
			os.Getenv("GOPATH") + "/src/git.qianqiusoft.com/qianqiusoft/light-apiengine/light-apiengine.xml",
			"../git.qianqiusoft.com/qianqiusoft/light-apiengine/light-apiengine.xml",
			"./vendor/git.qianqiusoft.com/qianqiusoft/light-apiengine/light-apiengine.xml",
			"./config/light-apiengine.xml",
			"./light-apiengine.xml",
		},
	)
	fmt.Println("read:" + fileName)
	bytess, _ := ioutil.ReadFile(fileName)
	err := xml.Unmarshal(bytess, app)
	if err != nil {
		fmt.Println("xml.Unmarshal(bytess, &app) error " + err.Error())
		return nil
	}
	return &MyXmlApplication{xml: app, Dir: "api_qian_qiu"}
}

func GetTestApp() *MyXmlApplication {
	app := &XmlApplication{}
	fileName := GetRightFile(
		[]string{
			os.Getenv("GOPATH") + "/src/git.i2edu.net/paddy.he/i2-erp-swagger/test.xml",
			"../git.i2edu.net/paddy.he/i2-erp-swagger/test.xml",
			"./vendor/git.i2edu.net/paddy.he/i2-erp-swagger/test.xml",
			"./test.xml",
		},
	)
	fmt.Println("warning use test xml : " + fileName + " !!!!!")
	bytess, _ := ioutil.ReadFile(fileName)
	err := xml.Unmarshal(bytess, app)
	if err != nil {
		fmt.Println("xml.Unmarshal(bytess, &app) error " + err.Error())
		return nil
	}
	return &MyXmlApplication{xml: app, Dir: "test"}
}

func main() {
	apps := make([]*MyXmlApplication, 0, 2)
	apps = append(apps, GetApiLightApp())
	apps = append(apps, GetProjectApp())
	if apps[0] == nil {
		apps = []*MyXmlApplication{GetTestApp()}
	}
	for _, app := range apps {
		swaggerDefinitions := GetSwaggerDefinitions(app.xml.Beans, app.xml.Tables)
		for _, controller := range app.xml.Controllers.ControllerList {
			swagger := GetSwagger(controller, swaggerDefinitions)
			NameToSwaggerBytes[controller.Name], _ = swagger.MarshalJSON()
			http.Handle("/swagger/"+app.GetDir(controller)+"-"+controller.Name+"-swagger.json", &Bs{NameToSwaggerBytes[controller.Name]})
		}
	}

	http.Handle("/swagger/api/menu", NewMapBs(func(s string) []byte {
		var tagsJson []byte
		switch s {
		case "st":
			tagsJson, _ = json.Marshal(GetMenu(apps, "st"))
		case "t":
			tagsJson, _ = json.Marshal(GetMenu(apps, "t"))
		case "s":
			tagsJson, _ = json.Marshal(GetMenu(apps, "s"))
		default: //"st"
			tagsJson, _ = json.Marshal(GetMenu(apps, "st"))
		}
		return tagsJson
	}))
	http.Handle("/", http.FileServer(http.Dir(
		GetRightDir(
			[]string{
				os.Getenv("GOPATH") + "/src/git.i2edu.net/paddy.he/i2-erp-swagger/static",
				"../git.i2edu.net/paddy.he/i2-erp-swagger/static",
				"./vendor/git.i2edu.net/paddy.he/i2-erp-swagger/static",
				"./static",
			},
		),
	)))
	if os.Getenv("SWAGGER_NOT_AUTO_OPEN") != "yes" {
		go func() {
			time.Sleep(300)
			err := Open("http://localhost:8087/swagger/dist")
			if err != nil {
				fmt.Println(err.Error())
			}
		}()
	}
	fmt.Println("listen port 8087 for swagger")
	err := http.ListenAndServe(":8087", nil)
	if err != nil {
		fmt.Println(err.Error())
	}
}

var commands = map[string][]string{
	"windows": {"cmd", "/c", "start"},
	"darwin":  {"open"},
	"linux":   {"xdg-open"},
}
var Version = "0.1.0"

func Open(uri string) error {
	run, ok := commands[runtime.GOOS]
	if !ok {
		return fmt.Errorf("don't know how to open things on %s platform", runtime.GOOS)
	}
	cmd := exec.Command(run[0], append(run[1:], uri)...)
	return cmd.Start()
}

type Bs struct {
	V []byte
}

func (p *Bs) ServeHTTP(w http.ResponseWriter, r *http.Request) {
	_, err := w.Write(p.V)
	if err != nil {
		fmt.Println(err)
	}
}

type MapBs struct {
	v     map[string][]byte
	getBs func(string) []byte
}

func NewMapBs(f func(string) []byte) *MapBs {
	return &MapBs{
		v:     make(map[string][]byte),
		getBs: f,
	}
}
func (p *MapBs) ServeHTTP(w http.ResponseWriter, r *http.Request) {
	key := r.FormValue("key")
	c := p.v[key]
	if c == nil {
		c = p.getBs(key)
		p.v[key] = c
	}
	_, err := w.Write(c)
	if err != nil {
		fmt.Println(err)
	}
}
